#if 0

Copyright (c) 2018, Christoph Schied
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the Karlsruhe Institute of Technology nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#endif
#include "a-svgf.h"

extern int gradientDownsample;

static int gradient_res(int x) { return (x + gradientDownsample - 1) / gradientDownsample; }

void A_SVGF::render(Scene::SharedPtr& pScene, Samples::SharedPtr& pSamples, Fbo::SharedPtr& pSourceFBO, Fbo::SharedPtr& pTargetFBO)
{
    int w = pTargetFBO->getWidth(), h = pTargetFBO->getHeight();
    float sceneRadius = pScene->getRadius();

    create_diff_fbo();

    PROFILE(svgf_reimplementation)
    if (!mEnable) {
        mpRenderContext->blitFbo(pSourceFBO.get(), pTargetFBO.get(),
            glm::ivec4(0, 0, w, h),
            glm::ivec4(0, 0, w, h));

        return;
    }

    auto &cam = *pScene->getActiveCamera();

    glm::vec2 jitterOffset = glm::vec2(
        cam.gePrevJitterX() - cam.getJitterX(),
        cam.gePrevJitterY() - cam.getJitterY());

    auto &frame_curr = pSamples->frame(0);
    auto &frame_prev = pSamples->frame(1);

    auto gbuffer_curr = frame_curr.mGBufferFBO;
    auto gbuffer_prev = frame_prev.mGBufferFBO;

    mpRenderContext->pushState();

    {
        RenderContext::Viewport vp;
        vp.width = float(gradient_res(pTargetFBO->getWidth()));
        vp.height = float(gradient_res(pTargetFBO->getHeight()));
        mpRenderContext->setViewport(0, vp);

        {
            PROFILE(create_diff_image)
            auto &pg = mpPrgCreateGradientSamples;
            auto &cb = mpPrgCreateGradientSamplesCB;

            mpRenderContext->setFbo(mpDiffPing);

            cb->setTexture("tex_color_unfiltered", frame_curr.mAuxiliaryDataFBO->getColorTexture(AuxiliaryDataId::Irradiance).get(), nullptr);
            cb->setTexture("tex_color_unfiltered_prev", frame_prev.mAuxiliaryDataFBO->getColorTexture(AuxiliaryDataId::Irradiance).get(), nullptr);
            cb->setTexture("tex_gradient_samples", frame_curr.mGradientPositions.get(), nullptr);
            cb->setVariable("gradientDownsample", gradientDownsample);
            cb->setTexture("tex_z", gbuffer_curr->getColorTexture(GBufferId::LinearZ).get(), nullptr);
            cb->setTexture("tex_vbuf", gbuffer_curr->getColorTexture(GBufferId::VBuffer).get(), nullptr);

            cb->uploadToGPU();
            mpRenderContext->setUniformBuffer(pg->getProgram()->getActiveProgramVersion()->getUniformBufferBinding("PerImageCB"), cb);
            pg->execute(mpRenderContext.get());
        }

        {
            PROFILE(filter_diff_image)
            auto &pg = mpPrgAtrousGradient;
            auto &cb = mpPrgAtrousGradientCB;

            for (int i = 0; i < mDiffAtrousIterations; i++) {
                mpRenderContext->setFbo(mpDiffPong);

                cb->setTexture("tex_color1", mpDiffPing->getColorTexture(0).get(), nullptr);
                cb->setTexture("tex_color2", mpDiffPing->getColorTexture(1).get(), nullptr);

                cb->setVariable("iteration", i);
                cb->setVariable("step_size", (1 << i));
                cb->setVariable("gradientDownsample", gradientDownsample);

                cb->uploadToGPU();
                mpRenderContext->setUniformBuffer(pg->getProgram()->getActiveProgramVersion()->getUniformBufferBinding("PerImageCB"), cb);
                pg->execute(mpRenderContext.get());

                std::swap(mpDiffPing, mpDiffPong);
            }

        }
    }


    RenderContext::Viewport vp;
    vp.width = float(pTargetFBO->getWidth());
    vp.height = float(pTargetFBO->getHeight());
    mpRenderContext->setViewport(0, vp);

    { /* temporal accumulation */
        PROFILE(svgf_reimpl_temporal_accum)
            std::swap(mpFBOAccum, mpFBOAccumPrev);

        auto &pg = mpPrgTemporalAccumulation;
        auto &cb = mpPrgTemporalAccumulationCB;
        mpRenderContext->setFbo(mpFBOAccum);

        cb->setTexture("tex_color", pSourceFBO->getColorTexture(0).get(), nullptr);
        cb->setTexture("tex_color_prev", mpColorHistory->getColorTexture(0).get(), nullptr);
        cb->setTexture("tex_moments_prev", mpFBOAccumPrev->getColorTexture(ACCUM_MOMENTS).get(), nullptr);
        cb->setTexture("tex_history_length", mpFBOAccumPrev->getColorTexture(ACCUM_LENGTH).get(), nullptr);
        cb->setTexture("tex_motion", gbuffer_curr->getColorTexture(GBufferId::MotionVecs).get(), nullptr);
        cb->setTexture("tex_z_curr", gbuffer_curr->getColorTexture(GBufferId::LinearZ).get(), nullptr);
        cb->setTexture("tex_z_prev", gbuffer_prev->getColorTexture(GBufferId::LinearZ).get(), nullptr);
        cb->setTexture("tex_normal_curr", gbuffer_curr->getColorTexture(GBufferId::Normal).get(), nullptr);
        cb->setTexture("tex_normal_prev", gbuffer_prev->getColorTexture(GBufferId::Normal).get(), nullptr);
        cb->setTexture("tex_vbuf_curr", gbuffer_curr->getColorTexture(GBufferId::VBuffer).get(), nullptr);
        cb->setTexture("tex_vbuf_prev", gbuffer_prev->getColorTexture(GBufferId::VBuffer).get(), nullptr);
        cb->setVariable("jitter_offset", jitterOffset);
        cb->setVariable("temporal_alpha", mTemporalAlpha);

        cb->setVariable("gradientDownsample", gradientDownsample);
        cb->setVariable("gradientFilterRadius", mGradientFilterRadius);
        cb->setVariable("normalizeGradient", int(mNormalizeGradient));
        cb->setVariable("antilag_alpha_scale", mAntilagAlphaScale);
        cb->setVariable("antilag_alpha_func", mAntilagAlphaFunc);
        cb->setTexture("tex_gradient_samples", frame_curr.mGradientPositions.get(), nullptr);
        cb->setTexture("tex_diff_current", mpDiffPing->getColorTexture(0).get(), nullptr);
        cb->setTexture("img_antilag_alpha", mpAntilagAlpha->getColorTexture(0).get(), nullptr, true);

        if (mShowAntilagAlpha)
            pg->getProgram()->addDefine("SHOW_ANTILAG_ALPHA", "1");
        else
            pg->getProgram()->removeDefine("SHOW_ANTILAG_ALPHA");

        cb->uploadToGPU();
        mpRenderContext->setUniformBuffer(pg->getProgram()->getActiveProgramVersion()->getUniformBufferBinding("PerImageCB"), cb);
        pg->execute(mpRenderContext.get());
    }

    { /* estimate variance */
        PROFILE(svgf_reimpl_estimate_variance)
        auto &pg = mpPrgEstimateVariance;
        auto &cb = mpPrgEstimateVarianceCB;
        mpRenderContext->setFbo(mpFBOPing);

        char buffer[64];
        snprintf(buffer, sizeof buffer, "%d", mNumIterations);
        pg->getProgram()->addDefine("ATROUS_ITERATIONS", buffer);

        cb->setTexture("tex_color", mpFBOAccum->getColorTexture(ACCUM_COLOR).get(), nullptr);
        cb->setTexture("tex_moments", mpFBOAccum->getColorTexture(ACCUM_MOMENTS).get(), nullptr);
        cb->setTexture("tex_history_length", mpFBOAccum->getColorTexture(ACCUM_LENGTH).get(), nullptr);

        cb->setTexture("tex_color_unfiltered", pSourceFBO->getColorTexture(0).get(), nullptr);
        cb->setTexture("tex_z", gbuffer_curr->getColorTexture(GBufferId::LinearZ).get(), nullptr);
        cb->setTexture("tex_normal", gbuffer_curr->getColorTexture(GBufferId::Normal).get(), nullptr);
        cb->setTexture("tex_vbuf", gbuffer_curr->getColorTexture(GBufferId::VBuffer).get(), nullptr);

        cb->uploadToGPU();
        mpRenderContext->setUniformBuffer(pg->getProgram()->getActiveProgramVersion()->getUniformBufferBinding("PerImageCB"), cb);
        pg->execute(mpRenderContext.get());

    }

    mpRenderContext->blitFbo(pSourceFBO.get(), mpColorHistoryUnfiltered.get(),
        glm::ivec4(0, 0, w, h),
        glm::ivec4(0, 0, w, h));

    if (mNumIterations == 0) {
        mpRenderContext->blitFbo(mpFBOPing.get(), mpColorHistory.get(),
            glm::ivec4(0, 0, w, h),
            glm::ivec4(0, 0, w, h)
        );
    }

    { /* atrous */
        PROFILE(svgf_reimpl_atrous)
        auto &pg = mpPrgAtrous;
        auto &cb = mpPrgAtrousCB;

        char buf[128];
        snprintf(buf, sizeof buf, "%d", mFilterKernel);
        pg->getProgram()->addDefine("FILTER_KERNEL", buf);

        for (int i = 0; i < mNumIterations; i++) {
            if (i - 1 == mHistoryTap) {
                mpRenderContext->blitFbo(mpFBOPing.get(), mpColorHistory.get(),
                    glm::ivec4(0, 0, w, h),
                    glm::ivec4(0, 0, w, h)
                );
            }
            mpRenderContext->setFbo(mpFBOPong);

            cb->setTexture("tex_color", mpFBOPing->getColorTexture(0).get(), nullptr);
            cb->setTexture("tex_z", gbuffer_curr->getColorTexture(GBufferId::LinearZ).get(), nullptr);
            cb->setTexture("tex_normal", gbuffer_curr->getColorTexture(GBufferId::Normal).get(), nullptr);
            cb->setTexture("tex_albedo", gbuffer_curr->getColorTexture(GBufferId::Albedo).get(), nullptr);

            cb->setVariable("iteration", i);
            cb->setVariable("step_size", (1 << i));
            cb->setVariable("modulate_albedo", int(i == mNumIterations - 1 && mModulateAlbedo));

            cb->uploadToGPU();
            mpRenderContext->setUniformBuffer(pg->getProgram()->getActiveProgramVersion()->getUniformBufferBinding("PerImageCB"), cb);
            pg->execute(mpRenderContext.get());


            std::swap(mpFBOPing, mpFBOPong);
        }

    }

    mpRenderContext->blitFbo(mpFBOPing.get(), pTargetFBO.get(),
        glm::ivec4(0, 0, w, h),
        glm::ivec4(0, 0, w, h));

    if (mShowAntilagAlpha) {
        mpRenderContext->blitFbo(mpAntilagAlpha.get(), pTargetFBO.get(),
            glm::ivec4(0, 0, w, h),
            glm::ivec4(0, 0, w, h));
    }

    mpRenderContext->popState();
}

void A_SVGF::resetState()
{
    mpFBOAccum->clearColorTarget(0, vec4(0));
    mpFBOAccum->clearColorTarget(1, vec4(0));
    mpFBOAccum->clearColorTarget(2, vec4(0));

    mpFBOAccumPrev->clearColorTarget(0, vec4(0));
    mpFBOAccumPrev->clearColorTarget(1, vec4(0));
    mpFBOAccumPrev->clearColorTarget(2, vec4(0));

    mpFBOPing->clearColorTarget(0, vec4(0));
    mpFBOPong->clearColorTarget(0, vec4(0));

    mpColorHistory->clearColorTarget(0, vec4(0));
}

void A_SVGF::load()
{
    Program::DefineList defineList;
    defineList.add("ANTILAG");
    char buf[128];
    snprintf(buf, sizeof buf, "%d", mFilterKernel);
    defineList.add("FILTER_KERNEL", buf);
    
#define _SVGF_SHADER(a) \
    mpPrg##a = FullScreenPass::create(#a ".frag", defineList); \
    mpPrg##a##CB = UniformBuffer::create(mpPrg##a->getProgram()->getActiveProgramVersion().get(), "PerImageCB");

    _SVGF_SHADER(TemporalAccumulation)
    _SVGF_SHADER(EstimateVariance)
    _SVGF_SHADER(Atrous)
    _SVGF_SHADER(CreateGradientSamples)
    _SVGF_SHADER(AtrousGradient)
#undef _SVGF_SHADER
}

void A_SVGF::shutdown()
{
    mpFBOAccum     = nullptr;
    mpFBOAccumPrev = nullptr;

    mpFBOPing = nullptr;
    mpFBOPong = nullptr;

    mpColorHistory = nullptr;

#define _SVGF_SHADER(a) \
    mpPrg##a = nullptr; \
    mpPrg##a##CB = nullptr;

    _SVGF_SHADER(TemporalAccumulation)
    _SVGF_SHADER(EstimateVariance)
    _SVGF_SHADER(Atrous)
    _SVGF_SHADER(CreateGradientSamples)
    _SVGF_SHADER(AtrousGradient)
#undef _SVGF_SHADER

    mpColorHistoryUnfiltered = nullptr;
}

void A_SVGF::initUI()
{
    Gui::dropdown_list filterKernels{
        { 0, "A-Trous" },
        { 1, "Box 3x3" },
        { 2, "Box 5x5" },
        { 3, "Sparse" },
        { 4, "Box3x3 / Sparse" },
        { 5, "Box5x5 / Sparse" },
    };
    mpGui->addCheckBox("Enable",                 &mEnable,                mGuiGroupName);
    mpGui->addCheckBox("Modulate Albedo",        &mModulateAlbedo,        mGuiGroupName);
    mpGui->addFloatVar("Temporal Alpha",         &mTemporalAlpha,         mGuiGroupName);
    mpGui->addIntVar  ("# Iterations",           &mNumIterations,         mGuiGroupName, 0, 16, 1);
    mpGui->addIntVar  ("History Tap",            &mHistoryTap,            mGuiGroupName, -1, 16, 1);
    mpGui->addDropdown("Kernel", filterKernels,  &mFilterKernel,          mGuiGroupName);
    mpGui->addCheckBox("Show Antilag Alpha",     &mShowAntilagAlpha,      mGuiGroupName);
    mpGui->addIntVar  ("# Diff Iterations",      &mDiffAtrousIterations,  mGuiGroupName, 0, 16, 1);
    mpGui->addIntVar  ("Gradient Filter Radius", &mGradientFilterRadius,  mGuiGroupName, 0, 16, 1);
}

void A_SVGF::destroyUI()
{
    mpGui->removeVar("Enable",             mGuiGroupName);
    mpGui->removeVar("Modulate Albedo",    mGuiGroupName);
    mpGui->removeVar("Temporal Alpha",     mGuiGroupName);
    mpGui->removeVar("# Iterations",       mGuiGroupName);
    mpGui->removeVar("History Tap",        mGuiGroupName);
    mpGui->removeVar("Kernel",             mGuiGroupName);
    mpGui->removeVar("Show Antilag Alpha", mGuiGroupName);
    mpGui->removeVar("# Diff Iterations",  mGuiGroupName);
    mpGui->removeVar("Gradient Filter Radius", mGuiGroupName);
}

void A_SVGF::create_diff_fbo()
{
    int w = mpFBOAccum->getWidth(), h = mpFBOAccum->getHeight();

    if(mpDiffPing && mpDiffPing->getWidth() == gradient_res(w)
    && mpDiffPing->getHeight() == gradient_res(h))
    {
        return;
    }

    Falcor::ResourceFormat format_diff_fbo[2] = {
        Falcor::ResourceFormat::RGBA32Float,
        Falcor::ResourceFormat::RGBA32Float,
    };

    mpDiffPing = createSharedFBO(gradient_res(w), gradient_res(h), format_diff_fbo, 1, ARRAYSIZE(format_diff_fbo));
    mpDiffPong = createSharedFBO(gradient_res(w), gradient_res(h), format_diff_fbo, 1, ARRAYSIZE(format_diff_fbo));
}

void A_SVGF::resizeSwapChain(const glm::ivec2& imgSize)
{
    int w = imgSize.x;
    int h = imgSize.y;
    Falcor::ResourceFormat format[3] = {
        Falcor::ResourceFormat::RGBA16Float,
        Falcor::ResourceFormat::RG32Float,
        Falcor::ResourceFormat::R16Float,
    };

    Falcor::ResourceFormat format_atrous[1] = {
        Falcor::ResourceFormat::RGBA16Float,
    };

    mpFBOAccum     = createSharedFBO(w, h, format, 1, 3);
    mpFBOAccumPrev = createSharedFBO(w, h, format, 1, 3);

    mpFBOPing     = createSharedFBO(w, h, format_atrous, 1, ARRAYSIZE(format_atrous));
    mpFBOPong     = createSharedFBO(w, h, format_atrous, 1, ARRAYSIZE(format_atrous));

    mpColorHistory = createSharedFBO(w, h, format, 1, 1);


    Falcor::ResourceFormat format_color_history[2] = {
        Falcor::ResourceFormat::RGBA32Float,
    };

    mpColorHistoryUnfiltered = createSharedFBO(w, h, format_color_history, 1, 1);

    Falcor::ResourceFormat format_antilag_alpha[1] = {
        Falcor::ResourceFormat::RGBA8Unorm
    };

    mpAntilagAlpha = createSharedFBO(w, h, format_antilag_alpha, 1, 1);
}
